/*
 * Copyright (c) 2008, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
 * may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <stdint.h>
#ifndef WIN32
#include <libgen.h>
#endif
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <cctype>
#include <cstdio>
#include <iostream>
#include <cstring>
#include <string>
#include <cstdlib>
#include <new>

#include "../logic/defines.h"
#include "../logic/util.h"
#include "../logic/dict.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/score.h"
#include "../logic/game.h"

#ifdef WIN32
# define random rand
# define srandom srand
#endif

#ifdef WIN32
# define ANSI_U ""
# define ANSI_B ""
# define ANSI_BU ""
# define ANSI_E ""
#else
# define ANSI_U "\033[4m"
# define ANSI_B "\033[1m"
# define ANSI_BU "\033[1;4m"
# define ANSI_E "\033[m"
#endif

using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::string;
using namespace Spellathon;

static void	solveit(const char *letters, const char *dpath);
static void	playit(const char *dpath, bool precomp, int min, int max);

static void	print_game(const char *word, int wpl);
static void	play_loop(const char *w, const vector<string> words);
static int	wpl_calc(int w);

static void	help(void);
static void	instr(void);

bool	keepscore = true;

int
main(int argc, char *argv[])
{
	char	letters[MAX_LETRS+1];
	char	dictpath[PATH_MAX];
	char	*tok = 0;
	int	i, ch, dflag=0, sflag=0, pflag=0, cflag=0;
	int 	minsol = MIN_SOL, maxsol = MAX_SOL;

#ifndef WIN32
	chdir(dirname(argv[0]));
#endif

	if (argc < 2)
		help();
	while ((ch = getopt(argc, argv, "nhcid:pr:s:")) != -1) {
		switch (ch) {
		case 'h':
			help();
			break;

		case 'i':
			instr();
			break;

		case 'r':
			minsol = maxsol = 0;
			minsol = strtol(optarg, 0, 10);
			if (minsol > 0) tok = strchr(optarg, ',');
			if (tok)	maxsol = strtol(tok+1, 0, 10);
			if (!maxsol || maxsol - minsol < MGAP_SOL) {
				cerr << "Invalid solution range."
					" (Compiled limit: " 
				     << MMIN_SOL << "," << MMAX_SOL << ")\n";
				exit(1);
			}
			break;

		case 's':
			sflag++;
			strncpy(letters, optarg, MAX_LETRS);
			letters[MAX_LETRS] = 0;
			for (i = 0; i < MAX_LETRS; i++) {
				if (!isalpha(letters[i])) {
					cerr << "Please enter seven letters.\n";
					exit(1);
				}
				letters[i] = tolower(letters[i]);
			}
			swapch(&letters[0], &letters[MAX_LETRS-1]);
			break;

		case 'p':
			pflag++;
			break;

		case 'c':
			cflag++;
			break;

		case 'd':
			dflag++;
			strncpy(dictpath, optarg, sizeof(dictpath)-1);
			dictpath[sizeof(dictpath)-1] = 0;
			break;

		case 'n':
			keepscore = false;
			break;

		case '?':
		default:
			help();
			break;
		}
	} 
	argc -= optind;
	argv += optind;

	if (!dflag) 
		strcpy(dictpath, DICTPATH);

	if (sflag && pflag) 
		help();

	if (sflag) 
		solveit(letters, dictpath);

	if (pflag) {
		srandom(time(0));
		playit(dictpath, cflag, minsol, maxsol);
	}

	return 0;
}

static void
playit(const char *dpath, bool precomp, int minsol, int maxsol)
{
	Dict dct(dpath, precomp);
	Game game(&dct);

	bool r = game.play(precomp, minsol, maxsol);
	if (r) 
		play_loop(game.sig().c_str(), game.words());
	else 
		cout <<  "No suitable game found.\n"
		       	 "Choose a different Solution range or use "
				 "a different dictionary.\n";
}
	
static void
play_loop(const char *w, const vector<string> words)
{
	int wpl = wpl_calc(words.size());
	DispWordsBuf dsp;
	dsp.init(words, wpl);
	time_t t = time(0);
	string op = "";
	int fw = wpl*9 - sizeof("words:   /  ");

	while (1) {
		print_game(w, wpl);

		if (op != "")
			op = "< " + op + " >";
		printf("%s\n", dsp.hintBufAnsiColor());
		printf(ANSI_B"%s%*swords: %d/%d\n\n"ANSI_E, op.c_str(), fw-(int)op.size()," ", 
			   dsp.nSetWords(), dsp.nWords());
		puts("Enter a word ('f' to finish, 'h' for hint):");

		string word;
		cin >> word;

		if (word == "h") {
			if (!dsp.setHint()) 
				op = "No more hints";
			else
				op = "";

			continue;
		}

		if (word == "f") {
fin:
			time_t t2 = time(0);

			print_game(w, wpl);
			printf("%s\n", dsp.solBufAnsiColor());
			if (keepscore) {
				printf(ANSI_B"Correct words: %d/%d    Time: %dm%ds\n"ANSI_E,
					dsp.nSetWords(), dsp.nWords(), 
					(int)(t2-t)/60, (int)(t2-5)%60);

				Score sc(dsp.nSetWords(), dsp.nWords(), 
					 dsp.hintLevel(), t2-t);

				printf(ANSI_B"Score: %d     Rating: %s\n\n"ANSI_E,
					 (int)sc.score(), sc.rating().c_str()); 
			} else {
				printf(ANSI_B"Finished in %d secs\n\n"ANSI_E, (int)(t2-t));
			}

			exit(0);
		}

		if (word.size() > 7 || word.size() < 4) {
				op = "No match";
		}

		switch (dsp.setWord(word)) {
		case 1:
			op = "Correct";
			if (dsp.nWords() == dsp.nSetWords()) 
				goto fin;			
			break;
		case -1:
			op = "No match";
			break;
		case 0:
			op = "Already entered";
			break;
		}
	}
}

static void
print_game(const char *word, int wpl)
{
	int lm = wpl * 9 / 2;
	if (lm >= 3)
		lm -= 3;

	printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n%*c\n\n", lm+3, toupper(word[1]));
	printf("%*c     %c\n\n", lm, toupper(word[2]), toupper(word[3]));
	printf("%*c"ANSI_BU"%c\n\n"ANSI_E, lm+2, ' ', toupper(word[6]));
	printf("%*c     %c\n\n", lm, toupper(word[4]), toupper(word[5]));
	printf("%*c\n\n", lm+3, toupper(word[0]));
}

static int
wpl_calc(int w)
{
	int wpl = w;

	if (wpl > 32) {
		wpl = (wpl+7) / 8;
		if (wpl > 8)
			wpl = 8;
	} else {
		wpl = 4;
	}

	return wpl;
}

static void
solveit(const char *letters, const char *dpath)
{

	Dict d(dpath, false);
	Game game(&d);
	DispWordsBuf dsp;

	int r = game.solve(letters);
	if (r) {
		dsp.init(game.words(), wpl_calc(game.words().size()));
		cout << endl << dsp.solBuf() << endl;
	} 
	cout << dsp.nWords() << " words\n\n";
}

static void	
help(void)
{
	cerr << "\nSpellathon "VERSION"    Copyright (c) 2008 Girish Rao\n\n"
		"spellathon < -p | -s <string> | -i > [ -r min,max ] [ -d /path/to/dict ] [ -c ] [ -n ]\n\n"
		"Main options:\n\n"
		"  -i  Game instructions.\n"
		"  -p  Play a new game.\n"
        "  -s  Solve a custom puzzle.\n"
		"      Enter the central letter followed by\n"
        "       the remaining letters.\n"
		"      Example: spellathon -s taplerss (central letter is 't').\n\n"
		"Sub options:\n\n"
		"  -r  Play games with #solutions in the range [min,max]\n"
		"  -c  Use precomputed dictionary.\n"
		"  -d  Use specified dictionary.\n"
		"  -n  No score.\n\n"
		;
	exit(1);
}

static void	
instr(void)
{
	printf("\n"ANSI_U"Game instructions"ANSI_E"\n\n"
		"  o Find all the words of four or more letters from the\n"
		"     letters shown in the puzzle.\n  o In making a word, each"
		" letter may be used once only.\n  o Each word must contain"
		" the central letter.\n  o There should be atleast one seven"
	        " letter word.\n\n"
		ANSI_U"Notes"ANSI_E"\n\n  - Using hints will reduce your score.\n"
		"  - To get the top score/rating you will need to be fast.\n"
		"  - Ratings: %s, %s, %s, %s, %s, %s.\n\n", Ratings[0], Ratings[1], 
		Ratings[2], Ratings[3], Ratings[4], Ratings[5]);
	exit(0);
}
