/*
 * Copyright (c) 2008, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
 * may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <QMainWindow>
#include <QMessageBox>
#include <QLabel>
#include <QSpinBox>
#include <QFileDialog>
#include <QDir>
#include <QCloseEvent>
#include <QSettings>
#include <QFont>
#include <QFontDialog>

#include "main_widget.h"
#include "puzzletrs_widget.h"
#include "../logic/defines.h"

MainWidget *gui_h;	/* global gui handle */

MainWidget::MainWidget(QWidget *parent)
	: QMainWindow(parent)
{
	gui_h = this;
	setupUi(this);
	spth = new SpellathonCls();
	settings = new QSettings("Girish", "Spellathon");

	puzzleLetrs_ = new PuzzleLetrs(puzzleBox);
	puzzleLetrs_->move((puzzleBox->size().width() - puzzleLetrs_->size().width())/2,
					   (puzzleBox->size().width() - puzzleLetrs_->size().width())/2);
	puzzleLetrs_->show();
	statWordCnt = new QLabel(this);
	statMsg = new QLabel(this);
	statWordCnt->setText("          ");
	statusOutput->insertWidget(0, statMsg, 1);
	statusOutput->insertPermanentWidget(1, statWordCnt, 0);
	playButton->setEnabled(false);
	solveButton->setEnabled(false);
	hintButton->setEnabled(false);
	finishButton->setEnabled(false);
	wordInput->setEnabled(false);

	settingsDialog = new QDialog(this);
	settingsUi.setupUi(settingsDialog);
	settingsUi.spinBox_max->setRange(MMIN_SOL + MGAP_SOL, MMAX_SOL);
	settingsUi.spinBox_min->setRange(MMIN_SOL, MMAX_SOL - MGAP_SOL);

	connect(action_Exit, SIGNAL(triggered()), 
			this, SLOT(close()));
	connect(action_Instructions, SIGNAL(triggered()), 
			this, SLOT(showInst()));
	connect(action_About, SIGNAL(triggered()), 
			this, SLOT(showAbout()));
	connect(action_Settings, SIGNAL(triggered()),
			this, SLOT(showSettings()));

	connect(exitButton, SIGNAL(clicked()), 
			this, SLOT(close()));
	connect(playButton, SIGNAL(clicked()), 
			this, SLOT(playPressed()));
	connect(hintButton, SIGNAL(clicked()), 
			wordInput, SLOT(setFocus()));
	connect(hintButton, SIGNAL(clicked()), 
			spth, SLOT(hint()));
	connect(finishButton, SIGNAL(clicked()), 
			this, SLOT(finishPressed()));
	connect(solveButton, SIGNAL(clicked()), 
			this, SLOT(solvePressed()));
	connect(wordInput, SIGNAL(returnPressed()), 
			this, SLOT(wordEntered()));
	connect(wordsOutput, SIGNAL(selectionChanged()),
			wordsOutput, SLOT(copy()));

	connect(spth, SIGNAL(printWords(QString)), 
			wordsOutput, SLOT(setText(QString)));
	connect(spth, SIGNAL(printWordsFmt(QString)), 
			wordsOutput, SLOT(setHtml(QString)));
	connect(spth, SIGNAL(printStatus(QString)), 
			statMsg, SLOT(setText(QString)));
	connect(spth, SIGNAL(printStatusTemp(QString, int)), 
			statusOutput, SLOT(showMessage(QString, int)));
	connect(spth, SIGNAL(printMsg(QString, QString)), 
			this, SLOT(printMsg(QString, QString)));
	connect(spth, SIGNAL(printWordCnt(QString)),
			statWordCnt, SLOT(setText(QString)));
	connect(spth, SIGNAL(fatalMsg(QString)), 
			this, SLOT(fatalMsg(QString)));
	connect(spth, SIGNAL(disableHints()), 
			this, SLOT(disableHints()));
	connect(spth, SIGNAL(printPuzzle(QString)), 
			this, SLOT(printPuzzle(QString)));
	connect(spth, SIGNAL(finished()), 
			this, SLOT(finishPressed()));
	connect(spth, SIGNAL(noDict(QString)), 
			this, SLOT(noDict(QString)));
	connect(spth, SIGNAL(spellathonReady()),
			this, SLOT(spellathonReady()));
	connect(spth, SIGNAL(wordchkimg(bool)),
			wordInput, SLOT(chk(bool)));

	connect(this, SIGNAL(play()), 
			spth, SLOT(play()));
	connect(this, SIGNAL(finish()), 
			spth, SLOT(finish()));
	connect(this, SIGNAL(setWord(QString)), 
			spth, SLOT(setWord(QString)));
	connect(this, SIGNAL(solve(QString)), 
			spth, SLOT(solve(QString)));
	connect(this, SIGNAL(setSolRange(int, int)), 
			spth, SLOT(setSolRange(int, int)));
	connect(this, SIGNAL(setDoScore(bool)),
			spth, SLOT(setDoScore(bool)));

	connect(settingsUi.dictSelectButton, SIGNAL(clicked()),
			this, SLOT(changeDict()));
	connect(settingsUi.monoFontButton, SIGNAL(clicked()),
			this, SLOT(changeFontMono()));
	connect(settingsDialog, SIGNAL(accepted()),
			this, SLOT(settingsAccepted()));
	connect(settingsUi.spinBox_max, SIGNAL(valueChanged(int)),
			this, SLOT(adjustSbMin(int)));
	connect(settingsUi.spinBox_min, SIGNAL(valueChanged(int)),
			this, SLOT(adjustSbMax(int)));

	for (int i = 0; i < 7; i++)
		connect(puzzleLetrs_->at(i), SIGNAL(textEdited(QString)), 
			SLOT(puzzleToUpper()));


	pendingDictChange = "";
	fontChanged = false;

	resize(settings->value("mainwidget/size", size()).toSize());
	move(settings->value("mainwidget/pos", pos()).toPoint());

	monoFont = settings->value("font/mono", wordsOutput->font()).value<QFont>();
	wordsOutput->setFont(monoFont);
	for (int i = 0; i < 7; i++) {
		puzzleLetrs_->at(i)->setFont(monoFont);
		puzzleLetrs_->at(i)->setContextMenuPolicy(Qt::NoContextMenu); 
	}
}

void
MainWidget::init()
{
	if (!settings->contains("dict/precomp"))
		settings->setValue("dict/precomp", PRECOMP);

	if (!settings->contains("mainwidget/size"))
		settings->setValue("mainwidget/size", size());
	if (!settings->contains("mainwidget/pos"))
		settings->setValue("mainwidget/pos", pos());
	QDir dp;
	if (!settings->contains("dict/path"))
		settings->setValue("dict/path", dp.absoluteFilePath(DICTPATH));
	if (!settings->contains("font/mono"))
		settings->setValue("font/mono", wordsOutput->font());
	if (!settings->contains("score/show"))
		settings->setValue("score/show", true);
	if (!settings->contains("solutions/min"))
		settings->setValue("solutions/min", MIN_SOL);
	if (!settings->contains("solutions/max"))
		settings->setValue("solutions/max", MAX_SOL);
	settings->sync();

	if (settings->status() != QSettings::NoError)
		fatalMsg(QString("Error updating settings: %1").arg(settings->fileName()));

	spth->init(settings->value("dict/path").toString(), 
			   settings->value("dict/precomp").toBool());

	emit setSolRange(settings->value("solutions/min").toInt(), 
					 settings->value("solutions/max").toInt());
	emit setDoScore(settings->value("score/show").toBool());
}

void 
MainWidget::adjustSbMin(int v)
{
	if (v - settingsUi.spinBox_min->value() >= MGAP_SOL)
		settingsUi.spinBox_min->setMaximum(v - MGAP_SOL);
}

void
MainWidget::changeFontMono()
{
	bool ok;

	QFont f = QFontDialog::getFont(&ok, wordsOutput->font(), settingsDialog);
	if (ok) {
		monoFont = f;
		fontChanged = true;
		settingsUi.monoFontLineEdit->setText(QString("%1, %2pt").arg(f.family()).arg(f.pointSize()));
	} else {
		fontChanged = false;
	}
}

void 
MainWidget::adjustSbMax(int v)
{
	if (settingsUi.spinBox_max->value() - v >= MGAP_SOL)
		settingsUi.spinBox_max->setMinimum(v + MGAP_SOL);
}

void 
MainWidget::showAbout()
{
	QMessageBox::about(this, "About", "Spellathon "VERSION
				 "\n\nCopyright (c) 2008 Girish Rao" 
				 "\n\nwww.sourceforge.net/projects/spellathon");
}

void 
MainWidget::showSettings()
{
	settingsUi.spinBox_min->setValue(settings->value("solutions/min").toInt());
	settingsUi.spinBox_max->setValue(settings->value("solutions/max").toInt());
	settingsUi.dictPathEdit->setText(settings->value("dict/path").toString());
	settingsUi.precompCheckBox->setChecked(settings->value("dict/precomp").toBool());
	settingsUi.scoreCheckBox->setChecked(settings->value("score/show").toBool());
	settingsDialog->resize(settings->value("settingswidget/size", 
						   settingsDialog->size()).toSize());
	QFont f = settings->value("font/mono").value<QFont>();
	settingsUi.monoFontLineEdit->setText(QString("%1, %2pt").arg(f.family()).arg(f.pointSize()));
	settingsDialog->exec();
}

void
MainWidget::changeDict()
{
	QString dpath = QFileDialog::getOpenFileName(settingsDialog, "Choose Dictionary", settings->value("dict/path").toString());

	if (dpath != "")
		settingsUi.dictPathEdit->setText(dpath);
}

void
MainWidget::settingsAccepted()
{
	emit setSolRange(settingsUi.spinBox_min->value(), 
			 settingsUi.spinBox_max->value());

	QString d =  settingsUi.dictPathEdit->text();	
	QDir dp;
	bool precomp = settingsUi.precompCheckBox->isChecked();

	settings->setValue("solutions/min", 
			   settingsUi.spinBox_min->value());
	settings->setValue("solutions/max", 
			   settingsUi.spinBox_max->value());

	settings->setValue("score/show",
			   settingsUi.scoreCheckBox->isChecked());
	emit setDoScore(settingsUi.scoreCheckBox->isChecked());

	if (fontChanged) {
		wordsOutput->setFont(monoFont); 
		for (int i = 0; i < 7; i++)
			puzzleLetrs_->at(i)->setFont(monoFont);
		settings->setValue("font/mono", monoFont);
	}

	settings->setValue("dict/path", d);
	settings->setValue("dict/precomp", precomp);

	if (d != spth->dict() || precomp != spth->isPrecomp()) {
		if (spth->isIngame())
			pendingDictChange = d;
		else
			spth->init(d, precomp);
	}

	settings->setValue("settingswidget/size", settingsDialog->size());
	settings->sync();
	if (settings->status() != QSettings::NoError)
		fatalMsg(QString("Error updating settings: %1").arg(settings->fileName()));
}

void 
MainWidget::showInst()
{
	QDialog *d = new QDialog(this);
	QTextEdit *t = new QTextEdit(d);
	QPushButton *p = new QPushButton(d);
	QVBoxLayout *l = new QVBoxLayout(d);

	connect(p, SIGNAL(clicked()), d, SLOT(close()));
	t->setLineWrapMode(QTextEdit::NoWrap);
	d->move(this->pos().x() + this->frameSize().width(), this->pos().y());
#ifdef WIN32
	d->resize(350, 480);
#else
	d->resize(380, 500);
#endif /* WIN32 */
	d->setWindowTitle("Instructions");
	t->setReadOnly(true);
	l->addWidget(t);
	l->addWidget(p, 0, Qt::AlignCenter);
	p->setText("&Ok");
	t->setLineWrapMode(QTextEdit::WidgetWidth);
	t->setHtml(
		QString("<br><b><u>Game instructions</u></b><br><br>"
		   "-&nbsp;Find all the words of four or more letters"
		   " from the letters shown in the puzzle.<br><br>"
		   "-&nbsp;In making a word, each letter may be used "
		   "once only.<br><br>-&nbsp;Each word must contain the "
		   "central letter.<br><br>-&nbsp;There should be atleast"
		   " one seven letter word.<br><br><br>"
		   "<b><u>Notes</u></b><br><br>"
		   "Hints will decrease your score.<br><br>"
		   "To get the top score/rating you will "
		   "also need to be fast.<br><br>"
		   "Ratings: %1, %2, %3, %4, %5, %6.<br><br>"
		   "To solve a custom puzzle enter the letters of puzzle "
		   "in the corresponding boxes and press Solve.<br>"
		   ).arg(Ratings[0]).arg(Ratings[1]).arg(Ratings[2])
			.arg(Ratings[3]).arg(Ratings[4]).arg(Ratings[5])
		);
	d->show();
}
	
void 
MainWidget::disableHints()
{
	hintButton->setEnabled(false);
	statusOutput->showMessage("No more hints.", 3000);
}

void
MainWidget::noDict(QString err)
{
	playButton->setEnabled(false);
	hintButton->setEnabled(false);
	solveButton->setEnabled(false);
	finishButton->setEnabled(false);
	for (int i = 0; i < 7; i++) {
		puzzleLetrs_->at(i)->clear();
		puzzleLetrs_->at(i)->setReadOnly(false);
		puzzleLetrs_->at(i)->setFocusPolicy(Qt::StrongFocus);
	}
	wordInput->clear();
	wordsOutput->clear();
	puzzleLetrs_->at(6)->setFocus();
	QMessageBox::critical(this, "Dictionary", err);
}

void
MainWidget::spellathonReady()
{
	playButton->setEnabled(true);
	solveButton->setEnabled(true);
	playButton->setFocus();
}

void 
MainWidget::playPressed()
{
	if (spth->isInprogress()) {
		int r = QMessageBox::warning(this, "New game", 
					     	"Game in progress. Quit current game?", 
					     	QMessageBox::Cancel | 
					     	QMessageBox::Ok, 
						QMessageBox::Cancel);

		if (r == QMessageBox::Cancel)
			return;
	}

	if (pendingDictChange != "") {
		spth->init(pendingDictChange, settings->value("dict/precomp").toBool());
		pendingDictChange = "";
		if (!spth->isDictOpen())
			return;
	}

	for (int i = 0; i < 7; i++) 
		puzzleLetrs_->at(i)->setText("");

	wordsOutput->setText("");
	wordInput->setEnabled(false);
	wordInput->clear();
	statMsg->clear();
	statWordCnt->setText("          ");
	playButton->setEnabled(false);
	hintButton->setEnabled(false);
	finishButton->setEnabled(false);
	solveButton->setEnabled(false);
	action_Settings->setEnabled(false);
	
	emit play();

	action_Settings->setEnabled(true);
	statMsg->clear();
	playButton->setEnabled(true);

	if (spth->isIngame()) {
		hintButton->setEnabled(true);
		finishButton->setEnabled(true);
		wordInput->setEnabled(true);
		wordInput->setFocus();
		wordInput->setInitState(true);
		for (int i = 0; i < 7; i++) { 
			puzzleLetrs_->at(i)->setReadOnly(true);
			puzzleLetrs_->at(i)->setFocusPolicy(Qt::NoFocus);
		}
	} else {
		for (int i = 0; i < 7; i++) { 
			puzzleLetrs_->at(i)->setReadOnly(false);
			puzzleLetrs_->at(i)->setFocusPolicy(Qt::StrongFocus);
		}

		solveButton->setEnabled(true);
	}
}

void 
MainWidget::fatalMsg(QString s)
{
	QMessageBox::critical(this, "Error", s);
}
	
void 
MainWidget::printMsg(QString t, QString s)
{
	QMessageBox::information(this, t, s);
}

void 
MainWidget::finishPressed()
{
	playButton->setEnabled(true);
	hintButton->setEnabled(false);
	finishButton->setEnabled(false);
	solveButton->setEnabled(true);
	playButton->setFocus();
	wordInput->clear();
	wordInput->setEnabled(false);

	for (int i = 0; i < 7; i++) {
		puzzleLetrs_->at(i)->setReadOnly(false);
		puzzleLetrs_->at(i)->setFocusPolicy(Qt::StrongFocus);
	}

	emit finish();
}

void
MainWidget::solvePressed()
{
	wordsOutput->setText("");
	wordInput->setEnabled(false);
	statWordCnt->setText("          ");
	statMsg->clear();

	if (pendingDictChange != "") {
		spth->init(pendingDictChange, settings->value("dict/precomp").toBool());
		pendingDictChange = "";
	}

	for (int i = 0; i < 7; i++) {
		if (!puzzleLetrs_->at(i)->isModified())
			goto nol;

		puzzle[i] = (puzzleLetrs_->at(i)->text())[0].toLower();
		if (!puzzle[i].isLower()) 
			goto nol;
	}

	emit solve(puzzle);
	puzzleLetrs_->at(6)->setFocus();
	return;

nol:
	for (int i = 0; i < 7; i++) 
		puzzleLetrs_->at(i)->clear();

	puzzleLetrs_->at(6)->setFocus();

	QMessageBox::information(this, "Solve puzzle", 
				 "Please enter the seven letters of the puzzle in the"
				 " corresponding boxes and then press Solve.");

}

void 
MainWidget::wordEntered()
{
	emit setWord(wordInput->text());
	wordInput->clear();
}

void 
MainWidget::puzzleToUpper()
{
	for (int i = 0; i < 7; i++) {
		puzzleLetrs_->at(i)->setText((puzzleLetrs_->at(i)->text()).toUpper());
		puzzleLetrs_->at(i)->setModified(true);
	}
}

void 
MainWidget::printPuzzle(QString s)
{
	QString::iterator it(s.begin()); 

	while(it != s.end()) { 
		*it = (*it).toUpper(); 
		it++;
	}

	for (int i = 0; i < 7; i++) {
		puzzleLetrs_->at(i)->setText(QString(s[i]));
		puzzleLetrs_->at(i)->setReadOnly(true);
	}

}

void
MainWidget::closeEvent(QCloseEvent *event)
{
	settings->setValue("mainwidget/size", size());
	settings->setValue("mainwidget/pos", pos());
	event->accept();
}

MainWidget::~MainWidget()
{
	delete spth;
	delete settings; 
}

