/*
 * Copyright (c) 2008, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
 * may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <new>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <errno.h>
#include <time.h>

#include "spellathon_cls.h"
#include "../logic/defines.h"
#include "../logic/dict.h"
#include "../logic/util.h"
#include "../logic/game.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/score.h"

namespace Spellathon
{
	
	SpellathonCls::SpellathonCls()
		:	dct(0), dsp(0), game(0),
			minSol_(MIN_SOL), maxSol_(MAX_SOL),
			ingame(false), inprogress(false),
			dictpath(""), dict_opened(false),
			precomp(false)
	{ }

	void
	SpellathonCls::setSolRange(int min, int max)
	{
		maxSol_ = max;
		minSol_ = min;
	}

	void 
	SpellathonCls::setDoScore(bool e)
	{
		doscore = e;
	}

	void
	SpellathonCls::init(QString dpath, bool pc)
	{
		destroy();

		dictpath = dpath;
		dict_opened = false;
		ingame = false;
		inprogress = false;
		precomp = pc;
		try {
			try {
				dct = new Dict(dictpath.toAscii().data(),
					       precomp);
			} 
			catch (string s) 
			{
				emit noDict(s.c_str());
				destroy();
				return;
			}
	
			try {	
				game = new Game(dct);
				dsp = new DispWordsBuf;
			} 
			catch (string s) 
			{
				emit fatalMsg(s.c_str());
				exit(1);
			}
		}
		catch (std::bad_alloc &ba)
		{
			string s = "Can't allocate memory: ";
		       	s += ba.what();
			emit fatalMsg(s.c_str());
			exit(1);
		}

		dict_opened = true;
		emit spellathonReady();	
	}
	
	void
	SpellathonCls::destroy()
	{
		delete dsp;		dsp = 0;
		delete game;	game = 0;
		delete dct;		dct = 0;
	}
	
	void
	SpellathonCls::play()
	{
		assert(dictpath != "");

		inprogress = false;
		
		bool r = game->play(precomp, minSol_, maxSol_);
		if (r) {
			ingame = true;
			t = time(0);
			emit printPuzzle(game->sig().c_str());
			dsp->init(game->words());
			emit printWords(dsp->hintBufHtml());
			emit printWordCnt(QString("Words:  0 / %1 ").arg(dsp->nWords()));
		} else {
			ingame = false;
			printMsg("No game", "No suitable game found.\n\n"
			       	 "Choose a different Solution range or use "
					 "a different dictionary.");
		}
	}
	
	void
	SpellathonCls::hint()
	{
		assert(ingame);
	
		dsp->setHint();
		emit printWords(dsp->hintBufHtml());
	
		if (dsp->hintLevel() >= 2)
			emit disableHints();
	}
	
	void
	SpellathonCls::finish()
	{
		assert(ingame);
	
		time_t t2 = time(0) - t;

		if (doscore) {
			Score sc(dsp->nSetWords(), dsp->nWords(), 
				 dsp->hintLevel(), t2);

			emit printStatus(QString("Time:  %1m%2s      Score:  %3")
					 .arg((int)t2/60)
					 .arg((int)t2%60).arg(sc.score()));

			emit printMsg("Score", 
				      QString("Correct words:  %1 / %2           Time: %5m %6s"
					      "\n\nScore:  %3    [%4]        \n")
				      .arg(dsp->nSetWords()).arg(dsp->nWords())
				      .arg(sc.score()).arg(sc.rating().c_str()).arg((int)t2/60).arg((int)t2%60));
		} else {
			emit printStatus(QString("Time:  %1m%2s")
					 .arg((int)t2/60).arg((int)t2%60));
		}

		emit printWordsFmt(dsp->solBufHtml());

		ingame = false;
		inprogress = false;
	}
	
	void
	SpellathonCls::solve(QString str)
	{
		assert(dictpath != "" && !ingame);

		int n = game->solve(str.toStdString());
		if (n > 0) {
			dsp->init(game->words());
			emit printWords(dsp->solBuf());
			emit printWordCnt(QString("Words:  %1 ").arg(dsp->nWords()));
		} else {
			emit printWordCnt("          ");
			emit printStatus("No Solution.");
		}
	}
	
	void
	SpellathonCls::setWord(QString w)
	{
		assert(ingame);
	
		switch (dsp->setWord(w.toStdString())) {
		case 1:
			inprogress = true;

			emit printWordCnt(QString("Words:  %1 / %2 ")
					  .arg(dsp->nSetWords())
					  .arg(dsp->nWords()));

			emit printWords(dsp->hintBufHtml());
			emit wordchkimg(true);

			if (dsp->nWords() == dsp->nSetWords()) 
				emit finished();
			
			break;

		case -1:
			emit wordchkimg(false);
			break;

		case 0:
			emit printStatusTemp("Already entered.", 3000);
			break;
		}
			
	}

}

