/*
 * Copyright (c) 2008, Girish Rao
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SPTH_CL
#include <QProgressBar>
#include <QCoreApplication>
#endif /* SPTH_CL */

#include "game.h"
#include "solver.h"
#include "dict.h"
#include "util.h"

#ifndef SPTH_CL
#include "../gui/main_widget.h"
extern MainWidget *gui_h;
#endif /* SPTH_CL */

#ifdef WIN32
/* mingw doesn't have random(3) */
# define random rand
# define srandom srand
#endif /* WIN32 */

namespace Spellathon
{
	Game::Game(Dict *dct) : dct_(dct), found_(false) 
	{
		sol_ = new Solver(dct_);
		dummy.push_back("");
		srandom(time(0));
	}

	Game::~Game()
	{
		delete sol_;
	}

	bool
	Game::play(bool precomp, int minsol, int maxsol)
	{
		if (precomp) 
			return _play_precomp(minsol, maxsol);
		else
			return _play(minsol, maxsol);
	}

	bool
	Game::_play(int minsol, int maxsol)
	{
		char		sig[MAX_LETRS+1];
		unsigned int 	(*sidx)[2];
		enum	{IDX, RAND};

	    sidx = (unsigned int (*)[2])malloc(2 * dct_->nent7() * sizeof(int));
		if (sidx == NULL) 
			spth_error(true, "Out of memory");

#ifndef SPTH_CL
		QProgressBar *pbar = new QProgressBar(gui_h->statusOutput);
		pbar->setRange(0, dct_->nent7());
		pbar->setTextVisible(false);
		pbar->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Ignored);
		gui_h->statusOutput->insertWidget(1, pbar, 0);
#endif
	
		/* shuffle the indexes. Knuth shuffle? */
		int j = dct_->off7();
		if (j < 0)
			goto notfound;

        for (int i = 0; i < dct_->nent7(); i++) {
                sidx[i][IDX] = j++;
                sidx[i][RAND] = random();
        }
	
        qsort((void*)sidx, dct_->nent7(), 2 * sizeof(int), sidxcmp);
	
		unsigned int cidx[MAX_LETRS][2];
        for (int i = 0; i < MAX_LETRS; i++) {
                cidx[i][IDX] = i;
                cidx[i][RAND] = random();
        }
	
        qsort((void*)cidx, MAX_LETRS, 2 * sizeof(int), sidxcmp);
	
		/*
		 * find a random seven letter signature from the dict and solve
		 * it and any of the six different central letter combinations.
		 * consider the word playable if MIN_SOL <= solutions <= MAX_SOL
		 *
		 * this technique works well. it's fast, sufficiently
		 * random and covers all the signatures.
		 */
		for (int i = 0; i < dct_->nent7(); i++) {

			const char *dsig = (*dct_).dp[sidx[i][IDX]].sign;

			int j;
			for (j = 0 ; j < MAX_LETRS; j++)
				sig[j] = dsig[cidx[j][IDX]]; 

			sig[j] = 0;

			int k = MAX_LETRS - 2;
			do {
				if (sol_->solve(sig, minsol, maxsol) != -1) {
					/* found suitable sign/game */
#ifndef SPTH_CL
					gui_h->statusOutput->removeWidget(pbar);
#endif /* SPTH_CL */
					free(sidx);
					sig_ = sig;
					return found_ = true;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
	
			} while (1);

			/* 
			 * this can take some time in pathological cases 
			 * so we need to keep the gui responsive 
			 */	
			if (!(i & ((1<<8) - 1))) {
#ifdef SPTH_CL
				spinner();
#else
				pbar->setValue(i);
				QCoreApplication::processEvents();
#endif /* SPTH_CL */
			}
		}
notfound:
		/* Uh Oh */	
#ifndef SPTH_CL
		gui_h->statusOutput->removeWidget(pbar);
#endif /* SPTH_CL */
		free(sidx);
		return found_ = false;
	}

	bool
	Game::_play_precomp(int minsol, int maxsol)
	{
		int i = 0, s = -1, e = -1;

		while (i < dct_->pc_nent()) {
			if (dct_->precomp_p[i].nsol >= minsol) {
				s = i;
				break;
			}
			i++;
		}

		while (i < dct_->pc_nent()) {
			if (dct_->precomp_p[i].nsol > maxsol) {
				break;
			}
			i++;
		}

		e = i;

		if (s < 0 || s >= e) {
			return found_ = false;
		} else {
			int r = randnum(s, e);
			sig_ = dct_->precomp_p[r].sign;
			sol_->solve(sig_.c_str());
			return found_ = true;
		}
	}

	bool
	Game::solve(string letters)
	{
		return found_ = sol_->solve(letters.c_str()) > 0;
	}

	const vector<string>& 
	Game::words() const
	{ 
		if (found_) 
			return sol_->words(); 
		else 
			return dummy; 
	}

	int 
	Game::sidxcmp(const void *a, const void *b)
	{
	        unsigned int (*c)[2] = (unsigned int (*)[2])a;
	        unsigned int (*d)[2] = (unsigned int (*)[2])b;
	
	        if ((*c)[1] == (*d)[1])
	                return 0;
	        if ((*c)[1] < (*d)[1])
	                return -1;
	        else
	                return 1;
	}
}

